import math
from typing import Tuple, Optional
from pyrpio.i2c_register_device import I2CRegisterDevice
from pyrpio.i2c import I2CBase


class LDC1X1Y:
    ''' TI LDC1X1Y base class. This class should be subclassed only. '''

    # LDC1X1Y register addresses (do not change)
    LDC1X1Y_DATA_BASE = 0x00
    LDC1X1Y_DATA_MSB_CH0 = 0x00
    LDC1X1Y_DATA_LSB_CH0 = 0x01
    LDC1X1Y_DATA_MSB_CH1 = 0x02
    LDC1X1Y_DATA_LSB_CH1 = 0x03
    LDC1X1Y_DATA_MSB_CH2 = 0x04
    LDC1X1Y_DATA_LSB_CH2 = 0x05
    LDC1X1Y_DATA_MSB_CH3 = 0x06
    LDC1X1Y_DATA_LSB_CH3 = 0x07

    LDC1X1Y_REF_COUNT_BASE = 0x08
    LDC1X1Y_REF_COUNT_CH0 = 0x08
    LDC1X1Y_REF_COUNT_CH1 = 0x09
    LDC1X1Y_REF_COUNT_CH2 = 0x0A
    LDC1X1Y_REF_COUNT_CH3 = 0x0B

    LDC1X1Y_OFFSET_BASE = 0x0C
    LDC1X1Y_OFFSET_CH0 = 0x0C
    LDC1X1Y_OFFSET_CH1 = 0x0D
    LDC1X1Y_OFFSET_CH2 = 0x0E
    LDC1X1Y_OFFSET_CH3 = 0x0F

    LDC1X1Y_SETTLE_COUNT_BASE = 0x10
    LDC1X1Y_SETTLE_COUNT_CH0 = 0x10
    LDC1X1Y_SETTLE_COUNT_CH1 = 0x11
    LDC1X1Y_SETTLE_COUNT_CH2 = 0x12
    LDC1X1Y_SETTLE_COUNT_CH3 = 0x13

    LDC1X1Y_CLOCK_DIVIDERS_BASE = 0x14
    LDC1X1Y_CLOCK_DIVIDERS_CH0 = 0x14
    LDC1X1Y_CLOCK_DIVIDERS_CH1 = 0x15
    LDC1X1Y_CLOCK_DIVIDERS_CH2 = 0x16
    LDC1X1Y_CLOCK_DIVIDERS_CH3 = 0x17

    LDC1X1Y_STATUS = 0x18
    LDC1X1Y_ERROR_CONFIG = 0x19
    LDC1X1Y_CONFIG = 0x1A
    LDC1X1Y_MUX_CONFIG = 0x1B
    LDC1X1Y_RESET_DEV = 0x1C

    LDC1X1Y_DRIVE_CURRENT_BASE = 0x1E
    LDC1X1Y_DRIVE_CURRENT_CH0 = 0x1E
    LDC1X1Y_DRIVE_CURRENT_CH1 = 0x1F
    LDC1X1Y_DRIVE_CURRENT_CH2 = 0x20
    LDC1X1Y_DRIVE_CURRENT_CH3 = 0x21

    LDC1X1Y_MANUFACTURER_ID = 0x7E
    LDC1X1Y_DEVICE_ID = 0x7F

    def __init__(self, bus: I2CBase, address=0x2A):
        self.address = address
        self.i2c_reg = I2CRegisterDevice(bus, address, register_size=1, data_size=2)

    def close(self):
        ''' Close up access. '''
        return 0

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    def get_register(self, register: int, mask: Optional[int] = None) -> int:
        ''' Get single word length register. '''
        value = self.i2c_reg.read_register(register)
        if mask is not None:
            value = value & mask
        return value

    def set_register(self, register, value: int, mask: Optional[int] = None):
        ''' Set single word length register. '''
        if mask is not None:
            pvalue = self.get_register(register, ~mask)  # pylint: disable=invalid-unary-operand-type
            value = pvalue | (value & mask)
        self.i2c_reg.write_register(register, value)

    def get_register_bit(self, register: int, bit: int):
        ''' Get single bit from register. '''
        mask = 1 << bit
        value = self.get_register(register, mask)
        return bool(value >> bit)

    def set_register_bit(self, register: int, bit: int, on: bool):
        ''' Set single bit of register. '''
        mask = 1 << bit
        pvalue = self.get_register(register, ~mask)
        self.set_register(register, pvalue | (int(on) << bit))

    def get_manufacturer_id(self) -> int:
        ''' Get manufacturer ID. '''
        return self.get_register(self.LDC1X1Y_MANUFACTURER_ID)

    def get_device_id(self):
        ''' Get device ID. '''
        return self.get_register(self.LDC1X1Y_DEVICE_ID)

    def get_channel_data(self, ch: int) -> Tuple[int, int]:
        ''' Get channel data and error flags. '''
        raise NotImplementedError()

    def get_channel_drive_current(self, ch: int) -> Tuple[int, int]:
        ''' Get channel L-C sensor drive current and computed sensor drive current
            Returns:
                int: Sensor drive current used during the settling + conversion time of channel.
                int: Initial and computed sensor drive current updated after each phase correction
        '''
        value = self.get_register(self.LDC1X1Y_DRIVE_CURRENT_BASE + ch)
        idrive = (value & 0xF800) >> 11
        init_idrive = (value & 0x07C0) >> 6
        return idrive, init_idrive

    def set_channel_drive_current(self, ch: int, value: int):
        ''' Set channel sensor drive current used during the settling +
            conversion time of channel.
            NOTE: RP_OVERRIDE_EN bit must be set to 1.
        '''
        self.set_register(self.LDC1X1Y_DRIVE_CURRENT_BASE + ch, (value & 0x001F) << 11)

    def get_channel_reference_count(self, ch: int) -> int:
        ''' Get channel reference clock count.
            0x0000-0x0004: Reserved
            0x0005-0xFFFF: Conversion Time (tC) = (RCOUNT x 16) / ƒREF
        '''
        return self.get_register(self.LDC1X1Y_REF_COUNT_BASE + ch)

    def set_channel_reference_count(self, ch: int, value: int):
        ''' Set channel reference clock count.
            0x0000-0x0004: Reserved
            0x0005-0xFFFF: Conversion Time (tC) = (RCOUNT x 16) / ƒREF
        '''
        if value <= 4:
            raise ValueError('Reference clock count must be > 4')
        return self.set_register(self.LDC1X1Y_REF_COUNT_BASE + ch, value)

    def get_channel_conversion_interval_time(self, ch: int, fref: float = 43.4e6) -> float:
        ''' Derive conversion interval time from ƒCLK and ƒREF
            Conversion Time (tC) = (RCOUNT × 16) / ƒREF
        '''
        return 16*self.get_channel_reference_count(ch=ch)/fref

    def get_channel_reference_offset(self, ch: int) -> int:
        ''' Get channel reference clock offset.
            fOFFSET = (OFFSET ÷ 2^16) × ƒREF
        '''
        return self.get_register(self.LDC1X1Y_OFFSET_BASE + ch)

    def set_channel_reference_offset(self, ch: int, value: int):
        ''' Set channel reference clock offset.
            ƒOFFSET = (OFFSET ÷ 2^16) × ƒREF
        '''
        return self.set_register(self.LDC1X1Y_OFFSET_BASE + ch, value)

    def get_channel_reference_settling_count(self, ch: int) -> int:
        ''' Get channel reference settling count.
            0x0000: Settle Time (tS)= 32 ÷ ƒREF
            0x0001: Settle Time (tS)= 32 ÷ ƒREF
            0x0002 - 0xFFFF: Settle Time (tS)= (SETTLECOUNT x 16) ÷ ƒREF
            NOTE: If the amplitude has not settled prior to the conversion start,
            an Amplitude error will be generated if reporting of this type of error is enabled.
        '''
        return self.get_register(self.LDC1X1Y_SETTLE_COUNT_BASE + ch)

    def set_channel_reference_settling_count(self, ch: int, value: int):
        ''' Set channel reference settling count.
            0x0000: Settle Time (tS) = 32 ÷ ƒREF
            0x0001: Settle Time (tS) = 32 ÷ ƒREF
            0x0002 - 0xFFFF: Settle Time (tS) = (SETTLECOUNT x 16) ÷ ƒREF
            NOTE: If the amplitude has not settled prior to the conversion start,
            an Amplitude error will be generated if reporting of this type of error is enabled.
        '''
        return self.set_register(self.LDC1X1Y_SETTLE_COUNT_BASE + ch, value)

    def get_channel_clock_dividers(self, ch: int) -> Tuple[int, int]:
        ''' Get clock dividers: ƒIN_DIV and ƒREF_DIV
            ƒIN_DIV must be set to ≥ 2 if the Sensor frequency is ≥ 8.75MHz
            ƒIN_DIV ≥ b0001:  ƒin0 = ƒSENSOR/ƒIN_DIV
            ƒREF_DIV is used to scale the maximum conversion frequency.
            ƒREF_DIV ≥ 0x001:  ƒREF = ƒCLK/ƒREF_DIV
        '''
        reg = self.get_register(self.LDC1X1Y_CLOCK_DIVIDERS_BASE + ch)
        fin_dev = (reg & 0xF000) >> 12
        fref_dev = (reg & 0x03FF) >> 0
        return fin_dev, fref_dev

    def set_channel_clock_dividers(self, ch: int, fin_div: int, fref_div: int):
        ''' Get clock dividers: fIN_DIV and fREF_DIV
            ƒIN_DIV must be set to ≥ 2 if the Sensor frequency is ≥ 8.75MHz
            ƒIN_DIV ≥ b0001: ƒin0 = ƒSENSOR/ƒIN_DIV
            ƒREF_DIV is used to scale the maximum conversion frequency.
            ƒREF_DIV ≥ 0x001:
            ƒREF0 = ƒCLK/ƒREF_DIV
        '''
        value = ((fin_div & 0x000F) << 12) | (fref_div & 0x03FF)
        self.set_register(self.LDC1X1Y_CLOCK_DIVIDERS_BASE + ch, value)

    def get_status_errors(self):
        ''' Get errors reported in status register.
            0xC0 - Channel bits CH0:0b00, CH1:0b01, CH2:0b10, CH3:0b11
            0x20: Under range error bit
            0x10: Over range error bit
            0x08: Watchdog timeout error bit
            0x04: Amplitude high error bit
            0x02: Amplitude low error bit
            0x01: Zero crossing error bit
        '''
        value = self.get_register(self.LDC1X1Y_STATUS) >> 8
        return value

    def device_reset(self):
        ''' Resets entire device to default values. '''
        self.set_register_bit(self.LDC1X1Y_RESET_DEV, 15, True)

    @property
    def high_current_drive(self):
        ''' High Current Sensor Drive
            b0: The LDC will drive all channels with normal sensor current (1.5mA max).
            b1: The LDC will drive channel 0 with current >1.5mA.
            NOTE: This mode is not supported if AUTOSCAN_EN = b1 (multi-channel mode).
        '''
        return self.get_register_bit(self.LDC1X1Y_CONFIG, 6)

    @high_current_drive.setter
    def high_current_drive(self, enable: bool):
        self.set_register_bit(self.LDC1X1Y_CONFIG, 6, enable)

    @property
    def status_update_interrupt_enable(self):
        ''' Enables status updates triggering interrupt pin. '''
        return self.get_register_bit(self.LDC1X1Y_CONFIG, 7)

    @status_update_interrupt_enable.setter
    def status_update_interrupt_enable(self, enable: bool):
        ''' Enable/disable status updates triggering interrupt pin. '''
        self.set_register_bit(self.LDC1X1Y_CONFIG, 7, enable)

    @property
    def reference_clock_external(self):
        ''' Enable external reference clock instead of internal oscillator. '''
        return self.get_register_bit(self.LDC1X1Y_CONFIG, 9)

    @reference_clock_external.setter
    def reference_clock_external(self, enable: bool):
        ''' Enable external reference clock or use internal oscillator. '''
        self.set_register_bit(self.LDC1X1Y_CONFIG, 9, enable)

    @property
    def automatic_amplitude_correction(self):
        ''' Use automatic amplitude correction.
            NOTE: Disable automatic correction for high precision applications.
        '''
        return not self.get_register_bit(self.LDC1X1Y_CONFIG, 10)

    @automatic_amplitude_correction.setter
    def automatic_amplitude_correction(self, enable: bool):
        return self.set_register_bit(self.LDC1X1Y_CONFIG, 10, not enable)

    @property
    def low_power_activation_mode(self):
        ''' Use low power sense mode- the LDC uses the value programmed in
            DRIVE_CURRENTx during sensor activation to minimize power consumption.
        '''
        return self.get_register_bit(self.LDC1X1Y_CONFIG, 11)

    @low_power_activation_mode.setter
    def low_power_activation_mode(self, enable: bool):
        self.set_register_bit(self.LDC1X1Y_CONFIG, 11, enable)

    @property
    def current_override_enable(self):
        ''' Provides control over sensor current drive used during the conversion time
            for Ch. x, based on the programmed value in the IDRIVEx field
        '''
        return self.get_register_bit(self.LDC1X1Y_CONFIG, 12)

    @current_override_enable.setter
    def current_override_enable(self, enable: bool):
        self.set_register_bit(self.LDC1X1Y_CONFIG, 12, enable)

    @property
    def sleep_mode(self):
        ''' Sleep mode or active mode. '''
        return self.get_register_bit(self.LDC1X1Y_CONFIG, 13)

    @sleep_mode.setter
    def sleep_mode(self, enable: bool):
        self.set_register_bit(self.LDC1X1Y_CONFIG, 13, enable)

    @property
    def data_ready(self):
        ''' A new conversion result is ready.
            When in Single Channel Conversion, this indicates a single conversion
            is available. When in sequential mode, this indicates that a new
            conversion result for all active channels is now available.
        '''
        return self.get_register_bit(self.LDC1X1Y_STATUS, 6)

    def channel_data_ready(self, ch: int):
        ''' New channel conversion result is ready. '''
        reg = self.get_register(self.LDC1X1Y_STATUS)
        return bool(reg & 0x000F & (1 << (3-ch)))

    def configure_single_active_channel(self, ch: int):
        ''' Configure single active channel for continuous sampling. '''
        self.set_register(self.LDC1X1Y_CONFIG, ch << 14, 0xC000)
        self.set_register(self.LDC1X1Y_MUX_CONFIG, value=0, mask=0xE000)

    def configure_sequential_channels(self, num_chs: int):
        ''' Configure sequential channels for continuous sampling.
            b00: Ch0, Ch1
            b01: Ch0, Ch1, Ch2 (LDC1X14 only)
            b10: Ch0, Ch1, Ch2, Ch3 (LDC1X14 only)
            b11: Ch0, Ch1
            Args:
                num_chs (int): Number of sequential channels
        '''
        seq_ch_map = {2: 0b100, 3: 0b101, 4: 0b110}
        seq_val = seq_ch_map[num_chs]
        self.set_register(self.LDC1X1Y_MUX_CONFIG, value=seq_val, mask=0xE000)

    def get_deglitch_filter_value(self) -> int:
        ''' Input Deglitch Filter Bandwidth
            0b001: 1.0 MHz | 0b100: 3.3 MHz | 0b101: 10 MHz | 0b111: 33 MHz
            Returns:
                int: raw register value
        '''
        value = self.get_register(self.LDC1X1Y_MUX_CONFIG, mask=0x0007)
        return value

    def get_deglitch_filter_frequency(self) -> float:
        ''' Input Deglitch Filter Bandwidth
            0b001: 1.0 MHz | 0b100: 3.3 MHz | 0b101: 10 MHz | 0b111: 33 MHz
            Returns:
                float: frequency in Hz
        '''
        bwid_val = self.get_deglitch_filter_value()
        bwid_map = {0b001: 1.0e6, 0b100: 3.3e6, 0b101: 10e6, 0b111: 33e6}
        bwid_freq = bwid_map[bwid_val]
        return bwid_freq

    def set_deglitch_filter_value(self, value: int):
        ''' Input Deglitch Filter Bandwidth
            Select the lowest setting that exceeds the maximum
            sensor oscillation frequency.
            0b001: 1.0 MHz | 0b100: 3.3 MHz | 0b101: 10 MHz | 0b111: 33 MHz
            Args:
                value (int): raw register value
        '''
        self.set_register(self.LDC1X1Y_MUX_CONFIG, value, mask=0x007)

    def set_deglitch_filter_frequency(self, freq: float):
        ''' Input Deglitch Filter Bandwidth
            Select the lowest setting that exceeds the maximum
            sensor oscillation frequency.
            0b001: 1.0 MHz | 0b100: 3.3 MHz | 0b101: 10 MHz | 0b111: 33 MHz
            Args:
                freq (float): Bandwidth frequency
        '''
        bwid_map = {1.0e6: 0b001, 3.3e6: 0b100, 10e6: 0b101, 33e6: 0b111}
        min_dist = freq
        bwid_freq = None
        bwid_val = 0b111
        for mf, mv in bwid_map.items():
            dist = abs(freq - mf)
            if dist < min_dist:
                min_dist = dist
                bwid_val = mv
                bwid_freq = mf
        print(f'Using closest frequency: {bwid_freq}')
        self.set_deglitch_filter_value(bwid_val)

    def set_clock_dividers_from_lc(self, ch: int, inductance: float, capacitance: float):
        ''' Set appropriate clock dividers for channel based on LC. '''
        fsensor = 1/(2*math.pi*math.sqrt(inductance*capacitance))/1e6  # in MHz
        fin_div = round(fsensor / 8.75 + 1)
        if 4*fsensor < 40:
            fref_div = 2  # fref = 40 / 2
        else:
            fref_div = 4  # fref = 40 / 4
        self.set_channel_clock_dividers(ch=ch, fin_div=fin_div, fref_div=fref_div)



from typing import Tuple
from pyrpio.i2c import I2CBase
from pyrpiic.sensor.ldc1x1y import LDC1X1Y


class LDC161X(LDC1X1Y):
    ''' TI 28-bit LDC1612/LDC1614 inductive sensor API. '''

    def __init__(self, bus: I2CBase, address: int = 0x2A):
        # Manufacturing ID: 0x5449
        # Device ID: 0x3055
        super().__init__(bus, address=address)

    def get_channel_data(self, ch: int) -> Tuple[int, int]:
        ''' Get channel data and error flags.
            Returns:
                int: Channel computed conversion value
                int: Error code
                    0x8: Under range error bit
                    0x4: Over range error bit
                    0x2: Watchdog timeout error bit
                    0x1: Amplitude error bit
        '''
        ch_msb = self.get_register(self.LDC1X1Y_DATA_BASE + 2*ch)
        ch_lsb = self.get_register(self.LDC1X1Y_DATA_BASE + 2*ch + 1)
        value = ((0x0FFF & ch_msb) << 16) | ch_lsb
        err_code = (ch_msb & 0xF000) >> 12
        return value, err_code



import math
from typing import Tuple


def float2frac(x: float, error: float = 1e-9) -> Tuple[int, int]:
    """Convert floating point to closest fraction

    Args:
        x (float): Floating-point number
        error (float, optional): Error bound. Defaults to 1e-9.

    Returns:
        Tuple[int, int]: Numerator, Denominator
    """
    n = int(math.floor(x))
    x -= n
    if x < error:
        return (n, 1)
    if 1 - error < x:
        return (n+1, 1)

    # The lower fraction is 0/1
    lower_n = 0
    lower_d = 1
    # The upper fraction is 1/1
    upper_n = 1
    upper_d = 1
    while True:
        # The middle fraction is (lower_n + upper_n) / (lower_d + upper_d)
        middle_n = lower_n + upper_n
        middle_d = lower_d + upper_d
        # If x + error < middle
        if middle_d * (x + error) < middle_n:
            # middle is our new upper
            upper_n = middle_n
            upper_d = middle_d
        # Else If middle < x - error
        elif middle_n < (x - error) * middle_d:
            # middle is our new lower
            lower_n = middle_n
            lower_d = middle_d
        # Else middle is our best fraction
        else:
            return n * middle_d + middle_n, middle_d



import time
import functools
import logging
from typing import Callable, Optional, List, Tuple
from pyrpio.i2c_register_device import I2CRegisterDevice
from pyrpio.i2c import I2CBase

log = logging.getLogger(__name__)

OutputEnableSetFunc = Optional[Callable[[bool], None]]
OutputEnableGetFunc = Optional[Callable[[], bool]]


class PCA9685:
    MODE1_REG_ADDR = 0x00
    MODE2_REG_ADDR = 0x01
    SUB_ADDR1_REG_ADDR = 0x02
    SUB_ADDR2_REG_ADDR = 0x03
    SUB_ADDR3_REG_ADDR = 0x04
    ALL_CALL_ADDR_REG_ADDR = 0x05
    CHANNEL_BASE_ADDR = 0x06

    def __init__(self, bus: I2CBase, address=0x20):
        self.address = address
        self.bus = bus
        self.i2c_reg = I2CRegisterDevice(bus, address, register_size=1, data_size=1)
        self._output_enable_setter: OutputEnableSetFunc = None
        self._output_enable_getter: OutputEnableGetFunc = None

    def get_register(self, register: int, mask: Optional[int] = None) -> int:
        ''' Get single byte register w/ optional mask. '''
        return self.i2c_reg.read_register(register, mask=mask)

    def set_register(self, register, value: int, mask: Optional[int] = None):
        ''' Set single byte register w/ optional mask. '''
        self.i2c_reg.write_register(register, value, mask=mask)

    def get_register_bit(self, register: int, bit: int) -> bool:
        ''' Get single bit from register. '''
        mask = 1 << bit
        value = self.get_register(register, mask)
        return bool(value >> bit)

    def set_register_bit(self, register: int, bit: int, on: bool):
        ''' Set single bit of a register. '''
        mask = 1 << bit
        pvalue = self.get_register(register, ~mask)
        self.set_register(register, pvalue | (int(on) << bit))

    def set_register_bits(self, register: int, bits: List[int], ons: List[bool]):
        ''' Set single bit of a register. '''
        mask = functools.reduce(lambda a, b: a | (1 << b), bits, 0)
        value = functools.reduce(lambda a, b: a | b, (int(on) << bit for on, bit in zip(ons, bits)))
        pvalue = self.get_register(register, ~mask)
        self.set_register(register, pvalue | value)

    def _get_channel_base_address(self, ch: int) -> int:
        ''' Get channel base register address. '''
        return PCA9685.CHANNEL_BASE_ADDR + 4*ch

    def set_output_enable_setter(self, setter: OutputEnableSetFunc = None):
        ''' Provide callable to toggle physical oe pin. '''
        self._output_enable_setter = setter

    def set_output_enable_getter(self, getter: OutputEnableGetFunc = None):
        ''' Provide callable to get physical oe pin state. '''
        self._output_enable_getter = getter

    @property
    def output_enable(self) -> bool:
        ''' Get state of output enable pin.
            NOTE: Requires calling `set_output_enable_getter` otherwise false will be returned.
        '''
        if self._output_enable_getter:
            return self._output_enable_getter()
        return False

    @output_enable.setter
    def set_output_enable(self, enable: bool):
        ''' Set output enable pin state.
            NOTE: Requires calling `set_output_enable_setter` otherwise no action is performed.
        '''
        if self._output_enable_setter:
            self._output_enable_setter(enable)

    def soft_reset(self):
        ''' Perform soft reset by writing 0x06 to i2c address 0x00. '''
        self.bus.set_address(0x00)
        self.bus.write(data=bytes([0x06]))
        self.bus.set_address(self.address)

    @property
    def pwm_prescale(self):
        ''' Get PWM prescale value [min=0x03 max=0xFF].
            prescale = round(osc_clk / (4096 x update_rate)) - 1
        '''
        return self.get_register(register=0xFE)

    @pwm_prescale.setter
    def pwm_prescale(self, prescale: int):
        ''' Set PWM prescale value.
            prescale = round(osc_clk / (4096 x update_rate)) - 1
        '''
        prescale = min(max(3, prescale), 255)
        self.set_register(register=0xFE, value=prescale)

    @property
    def mode1(self) -> int:
        ''' Get mode1 register value. '''
        return self.get_register(register=PCA9685.MODE1_REG_ADDR)

    @mode1.setter
    def mode1(self, value: int):
        ''' Set mode1 register value. '''
        return self.set_register(register=PCA9685.MODE1_REG_ADDR, value=value)

    @property
    def mode2(self) -> int:
        ''' Get mode2 register value. '''
        return self.get_register(register=PCA9685.MODE2_REG_ADDR)

    @mode2.setter
    def mode2(self, value: int):
        ''' Set mode2 register value. '''
        return self.set_register(register=PCA9685.MODE2_REG_ADDR, value=value)

    @property
    def subaddr1(self) -> int:
        ''' Get subaddress1 register value. '''
        return self.get_register(register=PCA9685.SUB_ADDR1_REG_ADDR)

    @subaddr1.setter
    def subaddr1(self, value: int):
        ''' Set subaddress1 register value. '''
        return self.set_register(register=PCA9685.SUB_ADDR1_REG_ADDR, value=value)

    @property
    def subaddr2(self) -> int:
        ''' Get subaddress2 register value. '''
        return self.get_register(register=PCA9685.SUB_ADDR2_REG_ADDR)

    @subaddr2.setter
    def subaddr2(self, value: int):
        ''' Set subaddress2 register value. '''
        return self.set_register(register=PCA9685.SUB_ADDR2_REG_ADDR, value=value)

    @property
    def subaddr3(self) -> int:
        ''' Get subaddress3 register value. '''
        return self.get_register(register=PCA9685.SUB_ADDR3_REG_ADDR)

    @subaddr3.setter
    def subaddr3(self, value: int):
        ''' Set subaddress3 register value. '''
        return self.set_register(register=PCA9685.SUB_ADDR3_REG_ADDR, value=value)

    @property
    def allcalladdr(self) -> int:
        ''' Get all-call address register value. '''
        return self.get_register(register=PCA9685.ALL_CALL_ADDR_REG_ADDR)

    @allcalladdr.setter
    def allcalladdr(self, value: int):
        ''' Set all-call address register value. '''
        return self.set_register(register=PCA9685.ALL_CALL_ADDR_REG_ADDR, value=value)

    @property
    def logic_restart_state(self) -> bool:
        ''' Get logic restart value. '''
        return self.get_register_bit(register=PCA9685.MODE1_REG_ADDR, bit=7)

    def restart_logic(self):
        ''' Restart chip logic. '''
        self.set_register_bit(register=PCA9685.MODE1_REG_ADDR, bit=7, on=True)

    @property
    def autoincrement(self) -> bool:
        ''' Get autoincrement i2c addresses flag. '''
        return self.get_register_bit(register=PCA9685.MODE1_REG_ADDR, bit=5)

    @autoincrement.setter
    def autoincrement(self, on: bool):
        ''' Set autoincrement i2c addresses flag. '''
        self.set_register_bit(register=PCA9685.MODE1_REG_ADDR, bit=5, on=on)

    @property
    def sleep(self) -> bool:
        ''' Get chip sleep mode. '''
        return self.get_register_bit(register=PCA9685.MODE1_REG_ADDR, bit=4)

    @sleep.setter
    def sleep(self, on: bool):
        ''' Set chip sleep mode. '''
        self.set_register_bit(register=PCA9685.MODE1_REG_ADDR, bit=4, on=on)

    @property
    def use_external_clock(self) -> bool:
        ''' Get external clock source flag. '''
        return self.get_register_bit(register=PCA9685.MODE1_REG_ADDR, bit=6)

    @use_external_clock.setter
    def use_external_clock(self, enable: bool):
        ''' Set whether external clock source or internal is used. '''

        if self.use_external_clock == enable:
            return
        if not enable:
            self.soft_reset()
        elif self.sleep:
            self.set_register_bit(register=PCA9685.MODE1_REG_ADDR, bit=6, on=True)
        else:
            self.sleep = True
            self.set_register_bits(register=PCA9685.MODE1_REG_ADDR, bits=[4, 6], ons=[True, True])
            time.sleep(500e-6)
            self.sleep = False
            time.sleep(500e-6)

    @property
    def subaddr1_set(self) -> bool:
        ''' Get if chip responds to subaddress 1. '''
        return self.get_register_bit(register=PCA9685.MODE1_REG_ADDR, bit=3)

    @subaddr1_set.setter
    def subaddr1_set(self, on: bool):
        ''' Set if chip responds to subaddress 1. '''
        return self.set_register_bit(register=PCA9685.MODE1_REG_ADDR, bit=3, on=on)

    @property
    def subaddr2_set(self) -> bool:
        ''' Get if chip responds to subaddress 2. '''
        return self.get_register_bit(register=PCA9685.MODE1_REG_ADDR, bit=2)

    @subaddr2_set.setter
    def subaddr2_set(self, on: bool):
        ''' Set if chip responds to subaddress 2. '''
        return self.set_register_bit(register=PCA9685.MODE1_REG_ADDR, bit=2, on=on)

    @property
    def subaddr3_set(self) -> bool:
        ''' Get if chip responds to subaddress 3. '''
        return self.get_register_bit(register=PCA9685.MODE1_REG_ADDR, bit=1)

    @subaddr3_set.setter
    def subaddr3_set(self, on: bool):
        ''' Set if chip responds to subaddress 3. '''
        return self.set_register_bit(register=PCA9685.MODE1_REG_ADDR, bit=1, on=on)

    @property
    def allcalladdr_set(self) -> bool:
        ''' Get if chip responds to All Call address. '''
        return self.get_register_bit(register=PCA9685.MODE1_REG_ADDR, bit=0)

    @allcalladdr_set.setter
    def allcalladdr_set(self, on: bool):
        ''' Set if chip responds to All Call address. '''
        return self.set_register_bit(register=PCA9685.MODE1_REG_ADDR, bit=0, on=on)

    @property
    def output_invert(self) -> bool:
        ''' Get if output logic state is inverted. '''
        return self.get_register_bit(register=PCA9685.MODE2_REG_ADDR, bit=4)

    @output_invert.setter
    def output_invert(self, on: bool):
        ''' Set if output logic state is inverted. '''
        return self.set_register_bit(register=PCA9685.MODE2_REG_ADDR, bit=4, on=on)

    @property
    def output_change(self) -> bool:
        ''' Get if outputs change on STOP command [0] or ACK [1]. '''
        return self.get_register_bit(register=PCA9685.MODE2_REG_ADDR, bit=3)

    @output_change.setter
    def output_change(self, on: bool):
        ''' Set if outputs change on STOP command [0] or ACK [1]. '''
        return self.set_register_bit(register=PCA9685.MODE2_REG_ADDR, bit=3, on=on)

    @property
    def output_drive(self) -> int:
        ''' Get output drive mode: 0 -> open-drain, 1 -> totem pole. '''
        return int(self.get_register_bit(register=PCA9685.MODE2_REG_ADDR, bit=2))

    @output_drive.setter
    def output_drive(self, drive: int):
        ''' Set output drive mode: 0 -> open-drain, 1 -> totem pole. '''
        return self.set_register_bit(register=PCA9685.MODE2_REG_ADDR, bit=2, on=bool(drive))

    @property
    def output_not_enable_mode(self) -> int:
        ''' Get output not enable mode. LEDn output when !OE = 1
            00: LEDn = 0
            01: LEDn = 1 when OUTDRV = 1 else high-impedance
            1X: LEDn = high-impedance
        '''
        return self.get_register(register=PCA9685.MODE2_REG_ADDR, mask=0x3)

    @output_not_enable_mode.setter
    def output_not_enable_mode(self, mode: int):
        ''' Set output not enable mode. '''
        return self.set_register(register=PCA9685.MODE2_REG_ADDR, value=mode, mask=0x3)

    def get_pwm_frequency(self, ref_freq: float = 25E6) -> float:
        ''' Get PWM frequency based on prescale and ref freq (Hz). '''
        return ref_freq / (4096.0*(self.pwm_prescale + 1.0))

    def set_pwm_frequency(self, freq: float, ref_freq: float = 25E6):
        ''' Set PWM prescale based on desired frequency in Hz. '''
        self.pwm_prescale = int(ref_freq / 4096.0 / freq + 0.5)

    def get_channel_on_counter(self, ch: int) -> Tuple[int, bool]:
        ''' Set channel on counter [0 - 4095]. Can also fully turn on channel. '''
        ch_base_addr = self._get_channel_base_address(ch=ch)
        lsb = self.get_register(register=ch_base_addr+0)
        msb = self.get_register(register=ch_base_addr+1)
        value = ((msb & 0x0F) << 8) | (lsb & 0xFF)
        full_on = bool(msb & 0x10)
        return value, full_on

    def get_channel_off_counter(self, ch: int) -> Tuple[int, bool]:
        ''' Set channel on counter [0 - 4095]. Can also fully turn on channel. '''
        ch_base_addr = self._get_channel_base_address(ch=ch)
        lsb = self.get_register(register=ch_base_addr+2)
        msb = self.get_register(register=ch_base_addr+3)
        value = ((msb & 0x0F) << 8) | (lsb & 0xFF)
        full_on = bool(msb & 0x10)
        return value, full_on

    def set_channel_on_counter(self, ch: int, on: int, full_on: bool):
        ''' Set channel on counter [0 - 4095]. Can also fully turn on channel. '''
        ch_base_addr = self._get_channel_base_address(ch=ch)
        self.set_register(register=ch_base_addr+0, value=on & 0xFF)  # ON LSB
        self.set_register(register=ch_base_addr+1, value=((on >> 8) & 0x0F) | (int(full_on) << 4))  # ON MSB

    def set_channel_off_counter(self, ch: int, off: int, full_off: bool):
        ''' Set channel off counter [0 - 4095]. Can also fully turn off channel. '''
        ch_base_addr = self._get_channel_base_address(ch=ch)
        self.set_register(register=ch_base_addr+2, value=off & 0xFF)  # OFF LSB
        self.set_register(register=ch_base_addr+3, value=((off >> 8) & 0x0F) | (int(full_off) << 4))  # OFF MSB

    def set_channel_full_on(self, ch: int, on: bool):
        ''' Set channel fully on (ignores on and off counters). '''
        ch_base_addr = self._get_channel_base_address(ch=ch)
        self.set_register_bit(register=ch_base_addr+1, bit=4, on=on)

    def set_channel_full_off(self, ch: int, off: bool):
        ''' Set channel fully off (ignores on and off counters). '''
        ch_base_addr = self._get_channel_base_address(ch=ch)
        self.set_register_bit(register=ch_base_addr+3, bit=4, on=off)

    def set_channel_counters(self, ch: int, on: int, off: int, full_on: bool, full_off: bool):
        ''' Set channels on and off counter. '''
        self.set_channel_on_counter(ch=ch, on=on, full_on=full_on)
        self.set_channel_off_counter(ch=ch, off=off, full_off=full_off)

    def get_channel_pwm(self, ch: int) -> Tuple[float, float]:
        ''' Get channels PWM duty cycle and delay. '''
        delay_cycles, full_on = self.get_channel_on_counter(ch=ch)
        off_cycles, full_off = self.get_channel_off_counter(ch=ch)
        if delay_cycles + (off_cycles - delay_cycles + 1) >= 4096 or off_cycles < delay_cycles:
            off_cycles += 4096
        on_cycles = off_cycles - delay_cycles + 1
        log.debug(
            'delay=%i, on=%i (full=%i), off=%i (full=%i)',
            delay_cycles, on_cycles, full_on, off_cycles, full_off
        )
        duty_cycle = 1 if full_on else 0 if full_off else round(on_cycles/4096, 3)
        delay = round((delay_cycles+1)/4096, 3)
        return duty_cycle, delay

    def set_channel_pwm(self, ch: int, duty_cycle: float, delay: float = 0):
        ''' Set channel counters based on desired duty cycle and delay.
            Args:
                duty_cycle: PWM duty cycle fraction [0, 1]
                delay: Fractional delay [0, 1]
        '''
        duty_cycle = max(min(duty_cycle, 1), 0)
        delay = max(min(delay, 1), 0)
        full_on = abs(1 - duty_cycle) <= 1e-3
        full_off = duty_cycle <= 1e-3
        if delay is None:
            delay = 0
        delay_cycles = max(0, round(delay*4096) - 1)
        on_cycles = max(0, round(duty_cycle*4096))
        off_cycles = delay_cycles + on_cycles
        off_cycles = max(0, off_cycles - 1 if off_cycles <= 4096 else off_cycles - 4096)
        log.debug(
            'delay=%i, on=%i (full=%i), off=%i (full=%i)',
            delay_cycles, on_cycles, full_on, off_cycles, full_off
        )
        if delay_cycles == off_cycles and not full_on and not full_off:
            raise ValueError('on counter and off counter cant be same value')
        self.set_channel_on_counter(ch=ch, on=delay_cycles, full_on=full_on)
        self.set_channel_off_counter(ch=ch, off=off_cycles, full_off=full_off)

    def disable_channel(self, ch: int):
        ''' Disable channel by setting LED off bit high'''
        self.set_channel_full_off(ch=ch, off=True)



import math
from typing import Optional, Tuple
from bitarray import bitarray
from pyrpio.i2c import I2CBase
from pyrpio.i2c_register_device import I2CRegisterDevice
from .defs import SI570Registers


class SI570:

    def __init__(self, bus: I2CBase, address: int):
        self.address = address
        self.i2c_reg = I2CRegisterDevice(bus, address, register_size=1, data_size=1)

    def get_registers(self, reg_addr: Optional[int] = 0x07) -> SI570Registers:
        """Read registers from clock IC into dataclass

        Args:
            reg_addr (Optional[int], optional): Register base address. Defaults to 0x07.

        Returns:
            SI570Registers: SI570 Registers
        """
        regs = SI570Registers()
        if reg_addr is not None:
            regs.reg_addr = reg_addr
        # Read raw bytes from registers
        data_in = bytearray()
        for i in range(6):
            chunk = self.i2c_reg.read_register(reg_addr + i)
            data_in.append(chunk)
        # Extract values from data
        raw_bits = bitarray(''.join([format(b, '08b') for b in data_in]))
        # First 3 bits hs, Next 7 bits N1, Following 38 bits Frequency
        regs.hs_div = int(raw_bits[:3].to01(), 2) + 4
        regs.n1 = int(raw_bits[3:10].to01(), 2) + 1
        fxp_f_req = int(raw_bits[10:].to01(), 2)
        regs.f_req = float(fxp_f_req)/(2.**28)
        return regs

    def freq2reg(self, freq_hz: float) -> SI570Registers:
        """Convert frequency to register dataclass

        Args:
            freq_hz (float): Target frequency in HZ

        Raises:
            Exception: Throws exception if target frequency is not achievable

        Returns:
            SI570Registers: SI570 Registers
        """
        regs = SI570Registers()
        f_out = freq_hz  # *1.0E6
        # Fixed values
        f_xtal = 114.285*1.0E6
        # Step 1: Assume f_dco = 5 GHz  (=f_xtal * f_req) to determine hs_div, n1
        f_dco = 5.0*1E9
        out_div = math.floor(f_dco/f_out)
        min_dist = float("inf")
        min_hs_div = 0
        min_n1 = 0
        found = False
        for hs_div in [4, 5, 6, 7, 9, 11]:
            n1 = out_div/hs_div
            n1 = n1 if n1 == 1 else 2*round(n1/2)
            new_dist = abs(n1*hs_div - out_div)
            if new_dist < min_dist and n1 <= 128:
                min_dist = new_dist
                min_hs_div = hs_div
                min_n1 = n1
                found = True
            #  if
        #  for
        regs.hs_div = min_hs_div
        regs.n1 = int(min_n1)
        if not found:
            raise Exception("Frequency not achievable- Fdco must be 4.85-5.67 GHz")
        # Step 2: Determine f_dco and therefore f_req
        f_dco = f_out * out_div
        regs.f_req = float(f_out * regs.hs_div * regs.n1)/f_xtal
        if f_dco < 4.85E9 or f_dco > 5.67E9:
            raise Exception("Frequency not achievable- Fdco must be 4.85-5.67 GHz")
        return regs

    def regs2freq(self, regs: SI570Registers) -> float:
        """Convert register dataclass to frequency

        Args:
            regs (SI570Registers): SI570 Registers

        Returns:
            float: Frequency in Hz
        """
        f_xtal = 114.285*1.0E6
        freq_hz = float(f_xtal*regs.f_req)/float(regs.hs_div*regs.n1)
        # freq_hz /= 1E6
        return freq_hz

    def set_registers(self, regs: SI570Registers, nonvolatile=False):
        """Writes registers to clock IC

        Args:
            regs (SI570Registers): SI570 Registers
            nonvolatile (bool, optional): Write to nonvolatile memory. Defaults to False.
        """
        # Binarize data
        #  First 3 bits hs, Next 7 bits N1,
        #  Following 38 bits Frequency (freq is in 10.28 fixed-point format)
        #  Append all for 48 bits total]
        fxp_freq = int(regs.f_req*2.0**28)
        reg_data = bitarray(format(regs.hs_div-4, '03b') + format(regs.n1-1, '07b') + format(fxp_freq, '038b'))
        regs_data = [int(reg_data[i:i+8].to01(), 2) for i in range(0, len(reg_data), 8)]
        # Read current registers
        res_reg = self.i2c_reg.read_register(135)
        frz_reg = self.i2c_reg.read_register(137)
        # Freeze DCO
        self.i2c_reg.write_register(137, frz_reg ^ 0x10)
        # Write to all registers (No need to chunk since less than 8)
        for i, v in enumerate(regs_data):
            self.i2c_reg.write_register(regs.reg_addr+i, v)
        # Unfreeze DCO
        self.i2c_reg.write_register(137, frz_reg & 0xEF)
        # Set NewFreq - New Frequency bit
        self.i2c_reg.write_register(135, res_reg ^ 0x40)

    def set_frequency(self, freq_hz: float, reg_addr: int = 0x07, nonvolatile=False):
        """Set clock IC to target frequency

        Args:
            freq_hz (float): Target frequency in Hz
            reg_addr (int, optional): Register address. Defaults to 0x07.
            nonvolatile (bool, optional): Write to nonvolatile memory. Defaults to False.
        """
        regs = self.freq2reg(freq_hz)
        self.set_registers(regs, nonvolatile=nonvolatile)

    def get_frequency(self, reg_addr: Optional[int] = 0x07) -> Tuple[float, SI570Registers]:
        """Get frequency and registers

        Args:
            reg_addr (Optional[int], optional): [description]. Defaults to 0x07.

        Returns:
            [type]: [description]
        """
        regs = self.get_registers(reg_addr=reg_addr)
        freq_hz = self.regs2freq(regs)
        return freq_hz, regs



import abc


class Clock(abc.ABC):

    @abc.abstractmethod
    def get_frequency(self, **kwargs):
        """ Read frequency from chip. """
        raise NotImplementedError()

    @abc.abstractmethod
    def set_frequency(self, **kwargs):
        """ Write desired frequency to chip. """
        raise NotImplementedError()



import array
import time
from enum import Enum
from pyrpio.i2c_register_device import I2CRegisterDevice
from pyrpio.i2c import I2CBase


# I2C Address
HDC1080_ADDRESS = (0x40)    # 1000000
# Registers
HDC1080_TEMPERATURE_REGISTER = (0x00)
HDC1080_HUMIDITY_REGISTER = (0x01)
HDC1080_CONFIGURATION_REGISTER = (0x02)
HDC1080_MANUFACTURERID_REGISTER = (0xFE)
HDC1080_DEVICEID_REGISTER = (0xFF)
HDC1080_SERIALIDHIGH_REGISTER = (0xFB)
HDC1080_SERIALIDMID_REGISTER = (0xFC)
HDC1080_SERIALIDBOTTOM_REGISTER = (0xFD)

# Configuration Register Bits

HDC1080_CONFIG_RESET_BIT = (0x8000)
HDC1080_CONFIG_HEATER_ENABLE = (0x2000)
HDC1080_CONFIG_ACQUISITION_MODE = (0x1000)
HDC1080_CONFIG_BATTERY_STATUS = (0x0800)
HDC1080_CONFIG_TEMPERATURE_RESOLUTION = (0x0400)
HDC1080_CONFIG_HUMIDITY_RESOLUTION_HBIT = (0x0200)
HDC1080_CONFIG_HUMIDITY_RESOLUTION_LBIT = (0x0100)

HDC1080_CONFIG_TEMPERATURE_RESOLUTION_14BIT = (0x0000)
HDC1080_CONFIG_TEMPERATURE_RESOLUTION_11BIT = (0x0400)

HDC1080_CONFIG_HUMIDITY_RESOLUTION_14BIT = (0x0000)
HDC1080_CONFIG_HUMIDITY_RESOLUTION_11BIT = (0x0100)
HDC1080_CONFIG_HUMIDITY_RESOLUTION_8BIT = (0x0200)


class HDC1080:
    class TempResolution(Enum):
        fourteen = 0
        eleven = 1

    class HumidityResolution(Enum):
        fourteen = 0
        eleven = 1
        eight = 2
    i2c: I2CBase
    i2c_reg: I2CRegisterDevice
    address: int

    def __init__(self, bus: I2CBase, address=HDC1080_ADDRESS):
        self.address = address
        self.i2c = bus
        self.i2c_reg = I2CRegisterDevice(
            bus, address, register_size=1, data_size=2)
        time.sleep(0.015)  # 15ms startup time

    def configure(self):
        """ configure for acquisition mode """
        config = HDC1080_CONFIG_ACQUISITION_MODE
        self.i2c.set_address(self.address)
        self.i2c_reg.write_register_bytes(HDC1080_CONFIGURATION_REGISTER,
                                          bytes(bytearray([config >> 8, 0x00])))
        time.sleep(0.015)

    def read_temperature(self) -> float:
        ''' read temperature and return a float '''
        s = bytes(bytearray([HDC1080_TEMPERATURE_REGISTER]))
        self.i2c.set_address(self.address)
        self.i2c.write(s)
        time.sleep(0.0625)              # Required delay

        self.i2c.set_address(self.address)
        data = self.i2c.read(2)  # read 2 byte temperature data
        buf = array.array('B', data)
        #print ( "Temp: %f 0x%X %X" % (  ((((buf[0]<<8) + (buf[1]))/65536.0)*165.0 ) - 40.0   ,buf[0],buf[1] )  )

        # Convert the data
        temp = (buf[0] * 256) + buf[1]
        cTemp = (temp / 65536.0) * 165.0 - 40
        return cTemp

    def read_humidity(self) -> float:
        ''' read humidity and return a float '''
        time.sleep(0.015)               # Required delay
        s = bytes(bytearray([HDC1080_HUMIDITY_REGISTER]))
        self.i2c.set_address(self.address)
        self.i2c.write(s)
        time.sleep(0.0625)              # Required delay

        data = self.i2c.read(2)  # read 2 byte humidity data
        buf = array.array('B', data)
        #print ( "Humidity: %f 0x%X %X " % (  ((((buf[0]<<8) + (buf[1]))/65536.0)*100.0 ),  buf[0], buf[1] ) )
        humidity = (buf[0] * 256) + buf[1]
        humidity = (humidity / 65536.0) * 100.0
        return humidity

    def read_config_register(self) -> int:
        ''' read configuration register and return integer value '''
        s = bytes(bytearray([HDC1080_CONFIGURATION_REGISTER]))
        self.i2c.set_address(self.address)
        self.i2c.write(s)
        time.sleep(0.0625)              # Required delay

        data = self.i2c.read(2)  # read 2 byte config data

        buf = array.array('B', data)

        # print("register={} {}".format(buf[0], buf[1]))
        return buf[0]*256+buf[1]

    def turn_heater_on(self):
        ''' turn heater on '''
        config = self.read_config_register()
        config = config | HDC1080_CONFIG_HEATER_ENABLE
        s = [HDC1080_CONFIGURATION_REGISTER, config >> 8, 0x00]
        s2 = bytes(bytearray(s))
        self.i2c.set_address(self.address)
        self.i2c.write(s2)  # sending config register bytes
        time.sleep(0.015)               # Required delay

    def turn_heater_off(self):
        ''' turn heater off '''
        config = self.read_config_register()
        config = config & ~HDC1080_CONFIG_HEATER_ENABLE
        s = [HDC1080_CONFIGURATION_REGISTER, config >> 8, 0x00]
        s2 = bytes(bytearray(s))
        self.i2c.set_address(self.address)
        self.i2c.write(s2)  # sending config register bytes
        time.sleep(0.015)               # Required delay

    def set_humidity_resolution(self, resolution: HumidityResolution):
        ''' set humidity resolution [0 - 14bit, 1 - 11bit, 2 - 8bit ] '''
        config = self.read_config_register()
        config = (config & ~0x0300) | resolution.value
        s = [HDC1080_CONFIGURATION_REGISTER, config >> 8, 0x00]
        s2 = bytes(bytearray(s))
        self.i2c.set_address(self.address)
        self.i2c.write(s2)  # sending config register bytes
        time.sleep(0.015)               # Required delay

    def set_temperature_resolution(self, resolution: TempResolution):
        ''' set temperature resolution [0 - 14bit, 1 - 11bit ] '''
        config = self.read_config_register()
        config = (config & ~0x0400) | resolution.value

        s = [HDC1080_CONFIGURATION_REGISTER, config >> 8, 0x00]
        s2 = bytes(bytearray(s))
        self.i2c.set_address(self.address)
        self.i2c.write(s2)  # sending config register bytes
        time.sleep(0.015)               # Required delay

    def read_battery_status(self) -> bool:
        ''' get battery status (bool) '''
        config = self.read_config_register()
        config = config & ~ HDC1080_CONFIG_HEATER_ENABLE

        return bool(config == 0)

    def read_manufacturer_id(self) -> int:
        ''' get manufacturuer id (int) '''
        s = [HDC1080_MANUFACTURERID_REGISTER]  # temp
        s2 = bytes(bytearray(s))
        self.i2c.set_address(self.address)
        self.i2c.write(s2)
        time.sleep(0.0625)              # Required delay

        data = self.i2c.read(2)  # read 2 byte config data

        buf = array.array('B', data)
        return buf[0] * 256 + buf[1]

    def read_device_id(self) -> int:
        ''' get device id (int) '''
        s = [HDC1080_DEVICEID_REGISTER]  # temp
        s2 = bytes(bytearray(s))
        self.i2c.set_address(self.address)
        self.i2c.write(s2)
        time.sleep(0.0625)              # Required delay

        data = self.i2c.read(2)  # read 2 byte config data

        buf = array.array('B', data)
        return buf[0] * 256 + buf[1]

    def read_serial_number(self) -> int:
        ''' get device serial number (int) '''
        serialNumber = 0

        s = [HDC1080_SERIALIDHIGH_REGISTER]  # temp
        s2 = bytes(bytearray(s))
        self.i2c.set_address(self.address)
        self.i2c.write(s2)
        time.sleep(0.0625)              # Required delay
        data = self.i2c.read(2)  # read 2 byte config data
        buf = array.array('B', data)
        serialNumber = buf[0]*256 + buf[1]

        s = [HDC1080_SERIALIDMID_REGISTER]  # temp
        s2 = bytes(bytearray(s))
        self.i2c.write(s2)
        time.sleep(0.0625)              # Required delay
        data = self.i2c.read(2)  # read 2 byte config data)
        buf = array.array('B', data)
        serialNumber = serialNumber*256 + buf[0]*256 + buf[1]

        s = [HDC1080_SERIALIDBOTTOM_REGISTER]  # temp
        s2 = bytes(bytearray(s))
        self.i2c.write(s2)
        time.sleep(0.0625)              # Required delay
        data = self.i2c.read(2)  # read 2 byte config data
        buf = array.array('B', data)
        serialNumber = serialNumber*256 + buf[0]*256 + buf[1]

        return serialNumber



from pyrpio.i2c import I2CBase

from .eeprom import EEPROM


class M24C02(EEPROM):
    """ Wrapper for EEPROM M24C02. Uses data sheet https://www.mouser.com/datasheet/2/389/m24c01-r-954990.pdf
    """

    def __init__(self, bus: I2CBase, address: int):
        super().__init__(bus=bus, address=address)



""" clock dataclass definitions """
from dataclasses import dataclass
from enum import Enum


class ClockType(str, Enum):
    LMK61E2 = 'LMK61E2'
    SI570 = 'SI570'


class LMK61E2ClockMode(int, Enum):
    TRISTATE = 0
    LVPECL = 1
    LVDS = 2
    HCSL = 3


@dataclass
class SI570Registers:
    hs_div: int = 0
    n1: int = 0
    f_req: float = 0
    reg_addr: int = 0x07


@dataclass
class LMK61E2Registers:
    pll_d: int = 1
    dmc: int = 0
    meo: int = 3
    cp: int = 4
    ps: int = 2
    c3: int = 1
    frac_num: int = 0
    frac_den: int = 1
    int_div: int = 0
    out_div: int = 1
    odf: LMK61E2ClockMode = LMK61E2ClockMode.LVDS



from typing import Tuple
from pyrpio.i2c import I2CBase
from pyrpiic.sensor.ldc1x1y import LDC1X1Y


class LDC141X(LDC1X1Y):
    ''' TI 14-bit LDC1412/LDC1414 inductive sensor API. '''

    def __init__(self, bus: I2CBase, address: int = 0x2A):
        # Manufacturing ID: 0x5449
        # Device ID: 0x3054
        super().__init__(bus, address=address)

    def get_channel_data(self, ch: int) -> Tuple[int, int]:
        ''' Get channel data and error flags.
            Returns:
                int: Channel computed conversion value
                int: Error code
                    0x8: Under range error bit
                    0x4: Over range error bit
                    0x2: Watchdog timeout error bit
                    0x1: Amplitude error bit
        '''
        ch_msb = self.get_register(self.LDC1X1Y_DATA_BASE + 2*ch)
        value = 0x0FFF & ch_msb
        err_code = (ch_msb & 0xF000) >> 12
        return value, err_code

    def get_output_gain(self):
        ''' Get output gain control.
            ---------------------------------------
            | BITS | GAIN | SHIFT  | RES. | RANGE |
            |------+------+--------+------+-------|
            |  00  |   1  | 0 bits |  12  | 100%  |
            |  01  |   4  | 2 bits |  14  | 25%   |
            |  10  |   8  | 3 bits |  15  | 12.5% |
            |  11  |  16  | 4 bits |  16  | 6.25% |
            ---------------------------------------
            Returns:
                int: Gain register value
        '''
        return self.get_register(self.LDC1X1Y_RESET_DEV, mask=0x0600) >> 9

    def set_output_gain(self, value: int):
        ''' Set output gain control.
            ---------------------------------------
            | BITS | GAIN | SHIFT  | RES. | RANGE |
            |------+------+--------+------+-------|
            |  00  |   1  | 0 bits |  12  | 100%  |
            |  01  |   4  | 2 bits |  14  | 25%   |
            |  10  |   8  | 3 bits |  15  | 12.5% |
            |  11  |  16  | 4 bits |  16  | 6.25% |
            ---------------------------------------
            Args:
                value (int): Gain register value
        '''
        return self.set_register(self.LDC1X1Y_RESET_DEV, value << 9, mask=0x0600)



import time

from pyrpio.i2c import I2CBase


class EEPROMException(Exception):
    pass


class EEPROM:
    """ An i2c eeprom reader. Based upon https://www.i2cchip.com/pdfs/I2C_EEProm_Reading_and_Programming.pdf for
    description of how these IC chips work and typical values to use. This only works with a single address,
    so for multi-block / address chips you must create multiple instances for each address.
    :param bus: I2C Bus
    """

    def __init__(self,
            bus: I2CBase,
            address: int,
            pages: int = 16,
            pointer_bytes: int = 1,
            page_bytes: int = 16,
            write_time_ms: int = 5
        ):
        '''
        Args:
            bus (I2C): I2C bus of the eeprom
            address (int): I2C address of the eerpom
            pages (int, optional):  Page count of the eeprom. Defaults to 16.
            pointer_bytes (int, optional): How many bytes for register addressing. Defaults to 1.
            page_bytes (int, optional): How many bytes in a page. Defaults to 16.
            write_time_ms (int, optional): Delay time after writing. Defaults to 5.
        '''
        self.__bus = bus
        self.__i2c_address = address
        self.__pointer_bytes = pointer_bytes
        self.__page_bytes = page_bytes
        self.__max_bytes = pages * page_bytes
        self.__write_time = write_time_ms / 1000.0

    def read_byte(self, address: int) -> bytes:
        '''
        Read byte at memory address

        Args:
            address: Memory address to read

        Returns:
            bytes: A single byte at memory address
        '''
        if address > self.__max_bytes:
            raise OverflowError(f'Overflows memory max is {self.__max_bytes} Bytes')
        self.__bus.set_address(self.__i2c_address)
        return self.__bus.read_write(address.to_bytes(length=1, byteorder='big'))

    def write_byte(self, address: int, value: bytes):
        '''
        Write byte at memory address

        Args:
            address (int): memory address to write
            value (int): byte to write
        '''
        if address > self.__max_bytes:
            raise EEPROMException(f'Overflows memory max is {self.__max_bytes} Bytes')
        self.__bus.set_address(self.__i2c_address)
        self.__bus.write(
            address.to_bytes(length=self.__pointer_bytes, byteorder='big') +
            value
        )
        time.sleep(self.__write_time)

    def read_sequential_bytes(self, start_address: int, num_bytes: int) -> bytes:
        '''
        Read sequential bytes

        Args:
            start_address (int): start address to read
            num_bytes (int): number of bytes to read

        Returns:
            bytes: bytes read
        '''
        if start_address + num_bytes > self.__max_bytes:
            raise EEPROMException(f'Overflows memory max is {self.__max_bytes} Bytes')
        self.__bus.set_address(self.__i2c_address)
        return self.__bus.read_write(start_address.to_bytes(length=self.__pointer_bytes, byteorder='big'), num_bytes)

    def write_sequential_bytes(self, start_address: int, data: bytes):
        '''
        Write bytes sequentially with paging algorithm

        Args:
            start_address (int): start address to write
            data (bytes): data to write
        '''

        if start_address + len(data) > self.__max_bytes:
            raise EEPROMException(f'Overflows memory max is {self.__max_bytes} Bytes')
        self.__bus.set_address(self.__i2c_address)
        sent_data = 0

        while sent_data < len(data):
            current_start_address = start_address + sent_data
            next_page_aligned_address = (current_start_address + self.__page_bytes) & (~(self.__page_bytes - 1))
            data_to_send = min(next_page_aligned_address - current_start_address, len(data) - sent_data)
            self.__bus.write(current_start_address.to_bytes(
                length=self.__pointer_bytes, byteorder='big') + data[sent_data: sent_data + data_to_send])
            time.sleep(self.__write_time)
            sent_data += data_to_send

    def write_string(self, start_address: int, value: str, encoding: str = 'ascii'):
        '''
        Encode and write string to eeprom

        Args:
            start_address (int): address to start writing
            value (str): string with no null character to encode and write
            encoding (str, optional): What encoding to use for the string. Defaults to 'ascii'.
        '''
        if '\0' in value:
            index = value.index('\0')
            raise ValueError(f'Value contains null character at index: {index}')
        self.write_sequential_bytes(start_address, (value + '\0').encode(encoding))

    def read_string(self, start_address: int, encoding: str = 'ascii'):
        '''
        Read and decode string from eeprom. Keeps reading pages until it finds null-terminating string

        Args:
            start_address (int): [description]
            encoding (str, optional): [description]. Defaults to 'ascii'.

        Returns:
            [type]: [description]
        '''
        value = ""
        offset = 0
        while True:
            address = start_address + offset
            if address > self.__max_bytes:
                value = None
                raise EOFError('No null character found. Possibly unterminated string or uninitialized memory.')
            data = self.read_sequential_bytes(start_address + offset, self.__page_bytes)
            sub_value = data.decode(encoding, errors='ignore')
            if '\0' in sub_value:
                return value + sub_value[:sub_value.index('\0')]
            value += sub_value
            offset += self.__page_bytes

    def dump(self) -> bytes:
        '''
        Dump all data of the eeprom

        Returns:
            bytes: all eeproms block data
        '''
        return self.read_sequential_bytes(start_address=0x0, num_bytes=self.__max_bytes)

    def erase_all(self):
        '''
        Erase (set bytes to 0x00) all data on the eeprom
        '''
        self.write_sequential_bytes(start_address=0x0, data=bytes(self.__max_bytes))



from typing import Union
from enum import Enum
from pyrpio.i2c_register_device import I2CRegisterDevice
from pyrpio.i2c import I2CBase


class GPIODir(str, Enum):
    IN = 'IN'
    OUT = 'OUT'


class PCA9698:
    PORT0 = ['IO0_0', 'IO0_1', 'IO0_2', 'IO0_3', 'IO0_4', 'IO0_5', 'IO0_6', 'IO0_7']
    PORT1 = ['IO1_0', 'IO1_1', 'IO1_2', 'IO1_3', 'IO1_4', 'IO1_5', 'IO1_6', 'IO1_7']
    PORT2 = ['IO2_0', 'IO2_1', 'IO2_2', 'IO2_3', 'IO2_4', 'IO2_5', 'IO2_6', 'IO2_7']
    PORT3 = ['IO3_0', 'IO3_1', 'IO3_2', 'IO3_3', 'IO3_4', 'IO3_5', 'IO3_6', 'IO3_7']
    PORT4 = ['IO4_0', 'IO4_1', 'IO4_2', 'IO4_3', 'IO4_4', 'IO4_5', 'IO4_6', 'IO4_7']
    PCA9698_BASE_INPUT = 0x00
    PCA9698_PORT0_INPUT = 0x00
    PCA9698_PORT1_INPUT = 0x01
    PCA9698_PORT2_INPUT = 0x02
    PCA9698_PORT3_INPUT = 0x03
    PCA9698_PORT4_INPUT = 0x04
    PCA9698_BASE_OUTPUT = 0x08
    PCA9698_PORT0_OUTPUT = 0x08
    PCA9698_PORT1_OUTPUT = 0x09
    PCA9698_PORT2_OUTPUT = 0x0A
    PCA9698_PORT3_OUTPUT = 0x0B
    PCA9698_PORT4_OUTPUT = 0x0C
    PCA9698_BASE_POLARITY = 0x10
    PCA9698_PORT0_POLARITY = 0x10
    PCA9698_PORT1_POLARITY = 0x11
    PCA9698_PORT2_POLARITY = 0x12
    PCA9698_PORT3_POLARITY = 0x13
    PCA9698_PORT4_POLARITY = 0x14
    PCA9698_BASE_CONFIG = 0x18
    PCA9698_PORT0_CONFIG = 0x18
    PCA9698_PORT1_CONFIG = 0x19
    PCA9698_PORT2_CONFIG = 0x1A
    PCA9698_PORT3_CONFIG = 0x1B
    PCA9698_PORT4_CONFIG = 0x1C

    def __init__(self, bus: I2CBase, address: int = 0x20):
        self.address = address
        self.i2c_reg = I2CRegisterDevice(bus, address, register_size=1, data_size=1)

    def close(self):
        ''' Close up access. '''
        return 0

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    def get_register_bit(self, register: int, bit: int):
        ''' Get single bit from register. '''
        mask = 1 << bit
        value = self.i2c_reg.read_register(register, mask)
        return bool(value >> bit)

    def set_register_bit(self, register: int, bit: int, on: bool):
        ''' Set single bit of a register. '''
        mask = 1 << bit
        pvalue = self.i2c_reg.read_register(register, ~mask)
        self.i2c_reg.write_register(register, pvalue | (int(on) << bit))

    def get_port_index(self, gpio: Union[str, int]) -> int:
        ''' Get port index for given gpio. '''
        if isinstance(gpio, str):
            if gpio in PCA9698.PORT0:
                return 0
            if gpio in PCA9698.PORT1:
                return 1
            if gpio in PCA9698.PORT2:
                return 2
            if gpio in PCA9698.PORT3:
                return 3
            if gpio in PCA9698.PORT4:
                return 4
        elif isinstance(gpio, int):
            return int(gpio / 8)
        raise ValueError(f'GPIO {gpio} is not a valid value')

    def get_gpio_bit_position(self, gpio: Union[str, int]) -> int:
        ''' Get register bit position for given gpio. '''
        if isinstance(gpio, str):
            if gpio in PCA9698.PORT0:
                return PCA9698.PORT0.index(gpio)
            if gpio in PCA9698.PORT1:
                return PCA9698.PORT1.index(gpio)
            if gpio in PCA9698.PORT2:
                return PCA9698.PORT2.index(gpio)
            if gpio in PCA9698.PORT3:
                return PCA9698.PORT3.index(gpio)
            if gpio in PCA9698.PORT4:
                return PCA9698.PORT4.index(gpio)
        elif isinstance(gpio, int):
            return int(gpio % 8)
        raise ValueError(f'GPIO {gpio} is not a valid value')

    def get_gpio_direction(self, gpio: Union[str, int]) -> GPIODir:
        ''' Get GPIO direction as either in or out. '''
        port_index = self.get_port_index(gpio)
        gpio_bit = self.get_gpio_bit_position(gpio)
        value = self.get_register_bit(self.PCA9698_BASE_CONFIG + port_index, gpio_bit)
        return GPIODir.IN if value else GPIODir.OUT

    def set_gpio_direction(self, gpio: Union[str, int], gpio_dir: GPIODir):
        ''' Set GPIO direction as either in or out. '''
        port_index = self.get_port_index(gpio)
        gpio_bit = self.get_gpio_bit_position(gpio)
        value = gpio_dir == GPIODir.IN
        self.set_register_bit(self.PCA9698_BASE_CONFIG + port_index, gpio_bit, value)

    def get_gpio_input(self, gpio: Union[str, int]) -> bool:
        ''' Read GPIO input value.'''
        port_index = self.get_port_index(gpio)
        gpio_bit = self.get_gpio_bit_position(gpio)
        return self.get_register_bit(self.PCA9698_BASE_INPUT + port_index, gpio_bit)

    def get_gpio_output(self, gpio: Union[str, int]) -> bool:
        ''' Get currently set GPIO output value.'''
        port_index = self.get_port_index(gpio)
        gpio_bit = self.get_gpio_bit_position(gpio)
        return self.get_register_bit(self.PCA9698_BASE_OUTPUT + port_index, gpio_bit)

    def set_gpio_output(self, gpio: Union[str, int], high: Union[bool, int]):
        ''' Pull GPIO output either active high or low.'''
        port_index = self.get_port_index(gpio)
        gpio_bit = self.get_gpio_bit_position(gpio)
        self.set_register_bit(self.PCA9698_BASE_OUTPUT + port_index, gpio_bit, bool(high))

    def get_gpio_polarity(self, gpio: Union[str, int]):
        ''' Get GPIO polarity setting. '''
        port_index = self.get_port_index(gpio)
        gpio_bit = self.get_gpio_bit_position(gpio)
        return self.get_register_bit(self.PCA9698_BASE_POLARITY + port_index, gpio_bit)

    def set_gpio_polarity(self, gpio: Union[str, int], flipped: bool):
        ''' Set GPIO polarity setting as either normal or flipped. '''
        port_index = self.get_port_index(gpio)
        gpio_bit = self.get_gpio_bit_position(gpio)
        self.set_register_bit(self.PCA9698_BASE_POLARITY + port_index, gpio_bit, flipped)



import math
import time
from bitarray import bitarray
from pyrpio.i2c import I2CBase
from pyrpio.i2c_register_device import I2CRegisterDevice
from .defs import LMK61E2ClockMode, LMK61E2Registers
from .utils import float2frac


class LMK61E2:

    def __init__(self, bus: I2CBase, address: int):
        """ The LMK61E2 device is an ultra-low jitter PLLatinumTM programmable oscillator with a fractional-N
        frequency synthesizer with integrated VCO that generates commonly used reference clocks.

        Args:
            bus (I2CBase): I2C bus interface
            address (int): I2C address of chip
        """
        self.address = address
        self.i2c_reg = I2CRegisterDevice(bus, address, register_size=1, data_size=1)

    def get_registers(self) -> LMK61E2Registers:
        ''' Read registers from device. '''
        # Read registers
        data_in = bytearray()
        for i in range(11):
            data_in.append(self.i2c_reg.read_register(22 + i))
        # REGISTER 33
        mash_ctrl_byte = self.i2c_reg.read_register(33)
        mash_ctrl_bits = bitarray(format(mash_ctrl_byte, '08b'))
        # REGISTER 34
        pll_ctrl0_byte = self.i2c_reg.read_register(34)
        pll_ctrl0_bits = bitarray(format(pll_ctrl0_byte, '08b'))
        # REGISTER 35
        pll_ctrl1_byte = self.i2c_reg.read_register(35)
        pll_ctrl1_bits = bitarray(format(pll_ctrl1_byte, '08b'))
        # REGISTER 21
        diffctrl_byte = self.i2c_reg.read_register(21)
        diffctrl_bits = bitarray(format(diffctrl_byte, '08b'))
        # Extract values from data
        regs = LMK61E2Registers()
        regs.pll_d = int(pll_ctrl0_bits[2])
        regs.cp = int(pll_ctrl0_bits[4:8].to01(), 2)
        regs.ps = int(pll_ctrl1_bits[1:4].to01(), 2)
        regs.c3 = int(pll_ctrl1_bits[5])
        regs.dmc = int(mash_ctrl_bits[4:6].to01(), 2)
        regs.meo = int(mash_ctrl_bits[6:8].to01(), 2)
        regs.odf = LMK61E2ClockMode(int(diffctrl_bits[6:8].to01(), 2))
        raw_bits = bitarray(''.join([format(b, '08b') for b in data_in]))
        regs.out_div = int(raw_bits[7:16].to01(), 2)
        regs.int_div = int(raw_bits[28:40].to01(), 2)
        regs.frac_num = int(raw_bits[42:64].to01(), 2)
        regs.frac_den = int(raw_bits[66:88].to01(), 2)
        return regs

    def regs2freq(self, regs: LMK61E2Registers) -> float:
        """Compute frequency (Hz) from registers.

        Args:
            regs (LMK61E2Registers): LMK61E2 register set

        Returns:
            float: Frequency in Hz
        """
        # Compute target frequency
        f_ref = 50.0*1E6
        frac_float = 0
        if regs.frac_den > 0 and regs.meo > 0:
            frac_float = regs.frac_num / regs.frac_den
        freq_hz = f_ref*(2*regs.pll_d)*(regs.int_div + frac_float)
        freq_hz = freq_hz / regs.out_div
        return freq_hz

    def freq2regs(self, freq_hz, odf: LMK61E2ClockMode = LMK61E2ClockMode.LVDS) -> LMK61E2Registers:
        """Compute registers from frequency (Hz). ODF is clock mode (2 = LVDS).

        Args:
            freq_hz (float): Target frequency in Hz
            odf (LMK61E2ClockMode, optional):  clock mode (2 = LVDS). Defaults to LMK61E2ClockMode.LVDS.

        Raises:
            Exception: If frequency is not achievable

        Returns:
            LMK61E2Registers: LMK61E2 registers
        """
        regs = LMK61E2Registers()
        f_out = freq_hz  # *1.0E6
        # Fixed values (Enable doubler by default)
        f_ref = 50.0*1.0E6
        regs.odf = LMK61E2ClockMode.LVDS.value
        regs.pll_d = 1
        # Step 1: Assume f_vco is in middle and determine out_div
        f_vco_middle = 5.0*1E9
        regs.out_div = int(round(f_vco_middle/f_out, 0))
        regs.out_div = max(min(regs.out_div, 511), 5)  # [5, 511]
        f_vco_desired = f_out * regs.out_div
        if f_vco_desired < 4.6E9 or f_vco_desired > 5.6E9:
            raise Exception('Frequency not achievable- Fvco must be in range [4.6 GHz, 5.6 GHz]')
        # Step 2: Assume fractional portion is just 0 to determine int_div
        regs.int_div = int(math.floor(f_vco_desired/(f_ref * (2*regs.pll_d)) - (0/1)))
        regs.int_div = max(min(regs.int_div, 4095), 1)
        f_vco_approx = f_ref * (2*regs.pll_d) * (regs.int_div + 0/1)
        # Step 3: If f_vco still off then adjust fractional portion
        if abs(f_vco_approx - f_vco_desired) < 1E-9:
            regs.frac_num = int(0)
            regs.frac_den = int(1)
            regs.ps = 0  # No phase shift
            regs.dmc = 3  # Disabled
            regs.meo = 0  # Integer mode
            regs.cp = 8  # 6.4 mA
            # NOTE: 3rd order filter should be ENABLED when doubler is enabled.
            # However, chip default has it disabled for int mode w/ doubler enabled
            # c3 = 1 if regs.pll_d is 1 else 0
            regs.c3 = 0
        else:
            frac_float = f_vco_desired/(f_ref * (2*regs.pll_d)) - regs.int_div
            frac_num, frac_den = float2frac(frac_float)
            regs.frac_num = int(frac_num)
            regs.frac_den = int(frac_den)
            regs.ps = 2  # 1 ns for 100 MHz fPD phase shift
            regs.dmc = 0  # Weak
            regs.meo = 3  # 3rd order
            regs.cp = 4  # 1.6 mA
            regs.c3 = 1  # 3rd order filter enabled
        return regs

    def set_registers(self, regs: LMK61E2Registers, nonvolatile: bool = False):
        """Writes registers to clock IC.

        Args:
            regs (LMK61E2Registers): LMK61E2 register set
            nonvolatile (bool, optional): Write to non-volatile memory. Defaults to False.
        """
        # Binarize data
        reg_data = bitarray(
            format(regs.out_div, '016b') +  # (7-bit 0 w/  9-bit OUTDIV)
            format(0, '08b') +  # (Skip register 24)
            format(regs.int_div, '016b') +  # (4-bit 0 w/ 12-bit INT)
            format(regs.frac_num, '024b') +  # (2-bit 0 w/ 22-bit NUM)
            format(regs.frac_den, '024b')  # (2-bit 0 w/ 22-bit DEN)
        )
        # Turn this into a byte array of bytes to send
        regs_pll_data = [int(reg_data[i:i+8].to01(), 2) for i in range(0, len(reg_data), 8)]

        # Write to main block of registers
        self.i2c_reg.write_register_sequential(22, regs_pll_data)
        # Write mash engine data
        mash_ctrl_bits = bitarray(format(regs.dmc, '06b') + format(regs.meo, '02b'))
        self.i2c_reg.write_register_sequential(33, [int(mash_ctrl_bits[0:8].to01(), 2)])
        # Write pll_d and cp register data
        pll_ctrl0_bits = bitarray(format(regs.pll_d, '03b') + format(regs.cp, '05b'))
        self.i2c_reg.write_register_sequential(34, [int(pll_ctrl0_bits[0:8].to01(), 2)])
        pll_ctrl1_bits = bitarray(format(regs.ps, '04b') + '0' + format(regs.c3, '01b') + '11')
        self.i2c_reg.write_register_sequential(35, [int(pll_ctrl1_bits[0:8].to01(), 2)])
        diffctrl_bits = bitarray(format(regs.odf, '08b'))
        self.i2c_reg.write_register_sequential(21, [int(diffctrl_bits[0:8].to01(), 2)])

        # Save register data to EEPROM (via SRAM)
        if nonvolatile:
            self.i2c_reg.write_register_sequential(49, [0x50])  # Copy regs to sram
            # Check register R49.6 bit to see when done (== 0)
            done = False
            while not done:
                nvmctrlByte = self.i2c_reg.read_register(49)
                nvmctrlBits = bitarray(format(nvmctrlByte, '08b'))
                done = int(nvmctrlBits[1]) == 0
                time.sleep(0.1)
            # Enable EEPROM write
            self.i2c_reg.write_register_sequential(56, [0xBE])
            # Perform EEPROM write
            self.i2c_reg.write_register_sequential(49, [0x11])
            time.sleep(0.1)
            done = False
            while not done:
                nvmctrlByte = self.i2c_reg.read_register(49)
                nvmctrlBits = bitarray(format(nvmctrlByte, '08b'))
                done = int(nvmctrlBits[5]) == 0
            # Disable EEPROM write
            self.i2c_reg.write_register_sequential(56, [0x00])

    def set_frequency(self, freq_hz: float, odf: LMK61E2ClockMode = LMK61E2ClockMode.LVDS,
                      nonvolatile: bool = False, **kwargs):
        ''' Set clock IC to target frequency '''
        regs = self.freq2regs(freq_hz, odf=odf)
        self.set_registers(regs, nonvolatile=nonvolatile)

    def get_frequency(self):
        ''' Get frequency from clock IC '''
        regs = self.get_registers()
        freq = self.regs2freq(regs)
        return freq, regs


# %%



import time
from typing import List, Union, Optional
from contextlib import contextmanager
from functools import reduce
from pyrpio.i2c import I2CBase

class PCA9543A:
    """The PCA9543A is a dual 2 bidirectional translating switch controlled by the I2C bus.
    """

    def __init__(self, bus: I2CBase, address: int = 0x70):
        """Constructor

        Args:
            bus: Subclass of I2CBase
            address: I2C address of device. Defaults to 0x70.
        """
        self.address = address
        self.bus = bus

    @property
    def num_lanes(self) -> int:
        """Get number of downstream lanes.

        Returns:
            int: number of downstream lanes
        """
        return 2

    @property
    def lanes(self) -> List[int]:
        """Get list of downstream lanes.

        Returns:
            List[int]: Downstream lanes
        """
        return [0, 1]

    def compute_value_from_lanes(self, lanes: Union[int, List[int]]) -> int:
        """Compute control register value from lanes.

        Args:
            lanes (Union[int, List[int]]): Target lanes

        Returns:
            int: Control register value
        """
        lanes = [lanes] if isinstance(lanes, int) else lanes
        return reduce(lambda acc, val: acc | (1 << val), lanes or [], 0)

    def get_control_register(self) -> int:
        """Get mux control register value.

        Returns:
            int: Control register
        """
        self.bus.set_address(self.address)
        return int.from_bytes(self.bus.read(length=1), byteorder='big')

    def set_control_register(self, value: int):
        """Set mux control register value.

        Args:
            value (int): Control register value
        """
        self.bus.set_address(self.address)
        self.bus.write(bytes([value]))

    def set_lanes(self, lanes: Union[int, List[int]]):
        """Set active downstream lanes.

        Args:
            lanes (Union[int, List[int]]): Downstream i2c lanes to enable
        """
        self.set_control_register(self.compute_value_from_lanes(lanes=lanes))
        time.sleep(0.05)

    def get_lanes(self) -> List[int]:
        """Get active downstream lanes.

        Returns:
            List[int]: Lanes
        """
        lanes = []
        value = self.get_control_register() & 0x3
        for lane in range(self.num_lanes):
            if value & (1 << lane):
                lanes.append(lane)
        return lanes

    def get_lane_interrupt(self, lane: int) -> bool:
        """Get interrupt flag for given lane.

        Args:
            lane (int): Downstream i2c lane

        Returns:
            bool: If interrupt flag is set
        """
        return bool(self.get_control_register() & (1 << (lane + 4)))

    def clear_lanes(self):
        """Clear all active downstream lanes.
        """
        self.set_control_register(value=0)

    @contextmanager
    def open_lanes(self, lanes: Optional[List[int]] = None, release: bool = False):
        """Open lanes only while in context block. Attempts to be re-entrant by only setting if active
        lanes are different and restores to what they were.
        Args:
            lanes (Optional[List[int]], optional): Downstream i2c lanes. Defaults to None.
            release (bool, optional): If True, will force release all lanes on exit. Defaults to False.

        Yields:
            int: Command register value
        """
        pvalue = 0
        value = self.compute_value_from_lanes(lanes)
        try:
            pvalue = self.get_control_register() & 0x3
            if pvalue != value:
                self.set_control_register(value=value)
            yield value
        finally:
            if release:
                self.clear_lanes()
            elif pvalue != value:
                self.set_control_register(value=pvalue)



from typing import Union
from enum import Enum
from pyrpio.i2c_register_device import I2CRegisterDevice
from pyrpio.i2c import I2CBase


class GPIODir(str, Enum):
    IN = "IN"
    OUT = "OUT"


class TCA6416A:
    PORT0 = ["P00", "P01", "P02", "P03", "P04", "P05", "P06", "P07"]
    PORT1 = ["P10", "P11", "P12", "P13", "P14", "P15", "P16", "P17"]
    TCA6416A_BASE_INPUT = 0x00
    TCA6416A_PORT0_INPUT = 0x00
    TCA6416A_PORT1_INPUT = 0x01
    TCA6416A_BASE_OUTPUT = 0x02
    TCA6416A_PORT0_OUTPUT = 0x02
    TCA6416A_PORT1_OUTPUT = 0x03
    TCA6416A_BASE_POLARITY = 0x04
    TCA6416A_PORT0_POLARITY = 0x04
    TCA6416A_PORT1_POLARITY = 0x05
    TCA6416A_BASE_CONFIG = 0x06
    TCA6416A_PORT0_CONFIG = 0x06
    TCA6416A_PORT1_CONFIG = 0x07

    def __init__(self, bus: I2CBase, address: int = 0x20):
        self.address = address
        self.i2c_reg = I2CRegisterDevice(bus, address, register_size=1, data_size=1)

    def close(self):
        """Close up access."""
        return 0

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    def get_register_bit(self, register: int, bit: int):
        """Get single bit from register."""
        mask = 1 << bit
        value = self.i2c_reg.read_register(register, mask)
        return bool(value >> bit)

    def set_register_bit(self, register: int, bit: int, on: bool):
        """Set single bit of a register."""
        mask = 1 << bit
        pvalue = self.i2c_reg.read_register(register, ~mask)
        self.i2c_reg.write_register(register, pvalue | (int(on) << bit))

    def get_port_index(self, gpio: str) -> int:
        """Get port index for given gpio."""
        if gpio in TCA6416A.PORT0:
            return 0
        if gpio in TCA6416A.PORT1:
            return 1
        raise ValueError(f"GPIO {gpio} is not a valid value")

    def get_gpio_bit_position(self, gpio: str) -> int:
        """Get register bit position for given gpio."""
        if gpio in TCA6416A.PORT0:
            return TCA6416A.PORT0.index(gpio)
        if gpio in TCA6416A.PORT1:
            return TCA6416A.PORT1.index(gpio)
        raise ValueError(f"GPIO {gpio} is not a valid value")

    def get_gpio_direction(self, gpio: str):
        """Get GPIO direction as either in or out."""
        port_index = self.get_port_index(gpio)
        gpio_bit = self.get_gpio_bit_position(gpio)
        value = self.get_register_bit(self.TCA6416A_BASE_CONFIG + port_index, gpio_bit)
        return GPIODir.IN if value else GPIODir.OUT

    def set_gpio_direction(self, gpio: str, gpio_dir: GPIODir):
        """Set GPIO direction as either in or out."""
        port_index = self.get_port_index(gpio)
        gpio_bit = self.get_gpio_bit_position(gpio)
        value = gpio_dir == GPIODir.IN
        self.set_register_bit(self.TCA6416A_BASE_CONFIG + port_index, gpio_bit, value)

    def get_gpio_input(self, gpio: str) -> bool:
        """Read GPIO input value."""
        port_index = self.get_port_index(gpio)
        gpio_bit = self.get_gpio_bit_position(gpio)
        return self.get_register_bit(self.TCA6416A_BASE_INPUT + port_index, gpio_bit)

    def get_gpio_output(self, gpio: str) -> bool:
        """Get currently set GPIO output value."""
        port_index = self.get_port_index(gpio)
        gpio_bit = self.get_gpio_bit_position(gpio)
        return self.get_register_bit(self.TCA6416A_BASE_OUTPUT + port_index, gpio_bit)

    def set_gpio_output(self, gpio: str, high: Union[bool, int]):
        """Pull GPIO output either active high or low."""
        port_index = self.get_port_index(gpio)
        gpio_bit = self.get_gpio_bit_position(gpio)
        self.set_register_bit(
            self.TCA6416A_BASE_OUTPUT + port_index, gpio_bit, bool(high)
        )

    def get_gpio_polarity(self, gpio: str):
        """Get GPIO polarity setting."""
        port_index = self.get_port_index(gpio)
        gpio_bit = self.get_gpio_bit_position(gpio)
        return self.get_register_bit(self.TCA6416A_BASE_POLARITY + port_index, gpio_bit)

    def set_gpio_polarity(self, gpio: str, flipped: bool):
        """Set GPIO polarity setting as either normal or flipped."""
        port_index = self.get_port_index(gpio)
        gpio_bit = self.get_gpio_bit_position(gpio)
        self.set_register_bit(
            self.TCA6416A_BASE_POLARITY + port_index, gpio_bit, flipped
        )

    def get_all_gpio_input(self, sequential=False) -> int:
        """Read GPIO input register pair"""
        if sequential:
            return self._get_gpio_register_pair_sequential(self.TCA6416A_BASE_INPUT)

        return self._get_gpio_register_pair(self.TCA6416A_BASE_INPUT)

    def set_all_gpio_output(self, value: int, sequential=False):
        """Set GPIO output register pair"""
        if sequential:
            self._set_gpio_register_pair_sequential(self.TCA6416A_BASE_OUTPUT, value)
        else:
            self._set_gpio_register_pair(self.TCA6416A_BASE_OUTPUT, value)

    def get_all_gpio_output(self, sequential=False) -> int:
        """Get GPIO output register pair"""
        if sequential:
            return self._get_gpio_register_pair_sequential(self.TCA6416A_BASE_OUTPUT)

        return self._get_gpio_register_pair(self.TCA6416A_BASE_OUTPUT)

    def set_all_gpio_polarity(self, value, sequential=False):
        """Set GPIO polarity register pair"""
        if sequential:
            self._set_gpio_register_pair_sequential(self.TCA6416A_BASE_POLARITY, value)
        else:
            self._set_gpio_register_pair(self.TCA6416A_BASE_POLARITY, value)

    def get_all_gpio_polarity(self, sequential=False) -> int:
        """Get GPIO polarity register pair"""
        if sequential:
            return self._get_gpio_register_pair_sequential(self.TCA6416A_BASE_POLARITY)

        return self._get_gpio_register_pair(self.TCA6416A_BASE_POLARITY)

    def set_all_gpio_direction(self, value: int, sequential=False):
        """Set GPIO configuration register pair"""
        if sequential:
            self._set_gpio_register_pair_sequential(self.TCA6416A_BASE_CONFIG, value)
        else:
            self._set_gpio_register_pair(self.TCA6416A_BASE_CONFIG, value)

    def get_all_gpio_direction(self, sequential=False) -> int:
        """Get GPIO configuration register pair"""
        if sequential:
            return self._get_gpio_register_pair_sequential(self.TCA6416A_BASE_CONFIG)

        return self._get_gpio_register_pair(self.TCA6416A_BASE_CONFIG)

    def _set_gpio_register_pair_sequential(self, base_addr: int, value: int):
        """Set GPIO register pair using sequential writes"""
        self.i2c_reg.write_register_sequential_bytes(
            base_addr, value.to_bytes(2, byteorder="little")
        )

    def _set_gpio_register_pair(self, base_addr: int, value: int):
        """Set GPIO register pair using two different write transactions"""
        self.i2c_reg.write_register(base_addr, (value >> 8) & 0xFF)
        self.i2c_reg.write_register(base_addr + 1, value & 0xFF)

    def _get_gpio_register_pair_sequential(self, base_addr: int) -> int:
        """Get GPIO register pair using sequential reads"""
        data = self.i2c_reg.read_register_sequential_bytes(base_addr, 2)
        return int.from_bytes(data, byteorder="little")

    def _get_gpio_register_pair(self, base_addr: int) -> int:
        """Get GPIO register pair using two different read transactions"""
        value = self.i2c_reg.read_register(base_addr)
        value |= self.i2c_reg.read_register(base_addr + 1) << 8
        return value
